from ._typing import AxesPermutation, Dims, Coords, BinaryOperator, Array, TernaryOperator, TensorLike
from ._typing import Function_2Args, Function_1Arg, Function_3Args, DimLike, DimsLike
from ._typing import HasDimName
